from typing import Optional

import torch
from torch import nn

from channel_attention.utils import create_conv_layer


class SEAttention(nn.Module):
    """
    The Squeeze-and-Excitation Attention for Time Series (1D) or Image (2D) Analysis.
    This module adaptively recalibrates channel-wise feature responses by explicitly modeling interdependencies between channels.

    Reference: "Squeeze-and-Excitation Networks" by Jie Hu, Li Shen, et al.

    URL: https://arxiv.org/abs/1709.01507
    """

    def __init__(
        self,
        n_dims: int,
        n_channels: int,
        reduction: Optional[int] = 4,
        bias: bool = False,
    ) -> None:
        """
        1D Squeeze-and-Excitation Attention for Time Series Analysis or
        2D Squeeze-and-Excitation Attention for Image Analysis.

        :param n_dims: (int) The dimension of input data, either 1 (time series) or 2 (image).
        :param n_channels: (int) The number of input channels of time series data.
        :param reduction: (int) The reduction ratio for the intermediate layer in the SE block.
        :param bias: (bool) Whether to include bias terms in the linear layers.
        """
        super().__init__()

        # Validate the input dimension
        assert n_dims in [1, 2], "The dimension of input data must be either 1 or 2."

        # The dimension of inputs data
        self.n_dims = n_dims

        # Global average pooling layer to squeeze the spatial dimensions
        self.avg_pool = (
            nn.AdaptiveAvgPool2d(1) if n_dims == 2 else nn.AdaptiveAvgPool1d(1)
        )

        # Fully connected layers for the excitation operation
        self.fc = nn.Sequential(
            nn.Linear(n_channels, n_channels // reduction, bias=bias),
            nn.ReLU(inplace=True),
            nn.Linear(n_channels // reduction, n_channels, bias=bias),
            nn.Sigmoid(),
        )

        # View shape for reshaping the excitation output
        self.view_shape = (1, 1) if n_dims == 2 else (1,)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the SEAttention module.

        :param x: (torch.Tensor)
                  1D Time Series: Input tensor of shape (batch_size, channels, seq_len);
                  2D Image: Input tensor of shape (batch_size, channels, height, width).

        :return: (torch.Tensor) Output tensor of the same shape as input
        """
        # Get the batch size, number of channels
        batch_size, channels = x.size()[:2]

        # Perform the Squeeze operation
        y = self.avg_pool(x).view(batch_size, channels)

        # Perform the Excitation operation
        y = self.fc(y).view(batch_size, channels, *self.view_shape)

        # Scale the input tensor with the recalibrated weights
        return x * y.expand_as(x)


class MultiSEAttention(nn.Module):
    """
    Multi-Branch Squeeze-and-Excitation Attention Module for Time Series (1D) or Image (2D) Analysis.
    This module enhances the representational power of the standard SE block by incorporating multiple branches and adaptive style assignment.
    """

    def __init__(
        self, n_dims: int, n_channels: int, reduction: int = 4, n_branches: int = 3
    ) -> None:
        """
        Multi-Branch Squeeze-and-Excitation Attention Module for Time Series (1D) or Image (2D) Analysis.

        :param n_dims: (int) The dimension of input data, either 1 (time series) or 2 (image).
        :param n_channels: (int) The number of input channels of time series data.
        :param reduction: (int) The reduction ratio for the intermediate layer in the SE block.
        :param n_branches: (int) The number of branches in the multi-branch SE module.
        """
        super(MultiSEAttention, self).__init__()

        # Dimension assertion
        assert n_dims in [1, 2], "The dimension of input data must be either 1 or 2."
        self.n_dims = n_dims

        # Create the average pooling layer and activation function
        self.avg_pool = (
            nn.AdaptiveAvgPool2d(1) if n_dims == 2 else nn.AdaptiveAvgPool1d(1)
        )
        self.activation = nn.Sigmoid()

        # Store the reduction ratio, number of branches, and number of channels
        self.reduction = reduction
        self.n_branches = n_branches
        self.n_channels = n_channels
        new_channels = n_channels * n_branches

        # Layers for multi-branch excitation
        self.fc = nn.Sequential(
            create_conv_layer(
                n_dims=n_dims,
                in_channels=new_channels,
                out_channels=new_channels // self.reduction,
                kernel_size=1,
                bias=True,
                groups=n_branches,
            ),
            nn.ReLU(inplace=True),
            create_conv_layer(
                n_dims=n_dims,
                in_channels=new_channels // self.reduction,
                out_channels=new_channels,
                kernel_size=1,
                bias=True,
                groups=n_branches,
            ),
        )

        # Style assignment layer
        self.style_assigner = nn.Linear(n_channels, n_branches, bias=False)

        # Repeat size for reshaping the output
        self.repeat_size = (1, 1) if n_dims == 2 else (1,)

    def _style_assignment(
        self, channel_mean: torch.Tensor, batch_size: int
    ) -> torch.Tensor:
        """
        Assign styles to each channel based on the channel mean.

        :param channel_mean: (torch.Tensor) The mean values of each channel, shape (batch_size, n_channels, 1, 1).
        :param batch_size: (int) The batch size of the input tensor.

        :return: (torch.Tensor) Style assignment probabilities for each branch, shape (batch_size, n_branches).
        """
        style_assignment = self.style_assigner(channel_mean.view(batch_size, -1))
        style_assignment = nn.functional.softmax(style_assignment, dim=1)
        return style_assignment

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the MultiSEAttention module.

        :param x: (torch.Tensor)
                  1D Time Series: Input tensor of shape (batch_size, channels, seq_len);
                  2D Image: Input tensor of shape (batch_size, channels, height, width).

        :return: (torch.Tensor) Output tensor of the same shape as input.
        """
        # Apply global average pooling
        avg_y = self.avg_pool(x)
        batch_size, n_channels = avg_y.shape[:2]

        # Perform style assignment
        style_assignment = self._style_assignment(avg_y, batch_size=batch_size)  # B x N

        # Multi-branch excitation
        avg_y = avg_y.repeat(1, self.n_branches, *self.repeat_size)

        # [batch_size, n_branches * n_channels, 1, 1]
        z = self.fc(avg_y)

        # Apply style assignment
        style_assignment = style_assignment.repeat_interleave(n_channels, dim=1)
        if self.n_dims == 1:
            z = z * style_assignment[:, :, None]
        else:
            z = z * style_assignment[:, :, None, None]

        # [batch_size, n_channels, 1, 1]
        z = torch.sum(
            z.view(batch_size, self.n_branches, n_channels, *self.repeat_size), dim=1
        )  # B x C x 1 x 1
        z = self.activation(z)

        return x * z



from typing import Optional

import torch
from torch import nn


def _transform_1d(data: torch.Tensor) -> torch.Tensor:
    """Transform 1D tensor by transposing the last two dimensions."""
    return data.transpose(-1, -2)


def _transform_2d(data: torch.Tensor) -> torch.Tensor:
    """Transform 2D tensor by squeezing the last dimension and transposing."""
    return data.squeeze(-1).transpose(-1, -2)


def _invert_transform_1d(data: torch.Tensor) -> torch.Tensor:
    """Invert transform for 1D tensor by transposing the last two dimensions."""
    return data.transpose(-1, -2)


def _invert_transform_2d(data: torch.Tensor) -> torch.Tensor:
    """Invert transform for 2D tensor by transposing and unsqueezing the last dimension."""
    return data.transpose(-1, -2).unsqueeze(-1)


class EfficientChannelAttention(nn.Module):
    """
    Efficient Channel Attention (ECA) module.
    This module captures cross-channel interaction efficiently without dimensionality reduction.

    Reference: ECA-Net: Efficient Channel Attention for Deep Convolutional Neural Networks

    URL: https://arxiv.org/abs/1910.03151
    """

    def __init__(
        self, n_dims: int, kernel_size: Optional[int] = 3, bias: Optional[bool] = False
    ) -> None:
        """
        Initialize EfficientChannelAttention module.

        :param n_dims: Number of dimensions of the input data (1 for time series, 2 for images).
        :param kernel_size: Size of the convolutional kernel (must be odd).
        :param bias: Whether to include a bias term in the convolutional layer.
        """
        super().__init__()
        # Check n_dims is valid for time series or image data
        assert n_dims in [
            1,
            2,
        ], "Only 1D and 2D `EfficientChannelAttention` are supported."
        self.n_dims = n_dims

        # Global average pooling layer to squeeze the spatial dimensions
        self.avg_pool = (
            nn.AdaptiveAvgPool2d(1) if n_dims == 2 else nn.AdaptiveAvgPool1d(1)
        )

        # Convolutional layer for capturing cross-channel interaction
        assert kernel_size % 2 == 1, "The kernel size must be odd."
        self.conv = nn.Conv1d(
            in_channels=1,
            out_channels=1,
            kernel_size=kernel_size,
            padding=(kernel_size - 1) // 2,
            bias=bias,
        )

        # Sigmoid activation function
        self.sigmoid = nn.Sigmoid()

        # Create the transformation functions
        if n_dims == 1:
            self.transform = _transform_1d
            self.invert_transform = _invert_transform_1d
        else:
            self.transform = _transform_2d
            self.invert_transform = _invert_transform_2d

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the EfficientChannelAttention module.

        :param x: (torch.Tensor) Input tensor of shape (batch_size, n_channels, length) for 1D
                  or (batch_size, n_channels, height, width) for 2D.

        :return: (torch.Tensor) Output tensor of the same shape as input.
        """
        # feature descriptor on the global spatial information
        y = self.avg_pool(x)

        # Two different branches of ECA module
        y = self.transform(y)
        y = self.conv(y)
        y = self.invert_transform(y)

        # Multi-scale information fusion
        y = self.sigmoid(y)

        return x * y.expand_as(x)



__all__ = [
    "SEAttention",
    "MultiSEAttention",
    "SpatialAttention",
    "ChannelAttention",
    "ConvBlockAttention",
    "EfficientChannelAttention",
    "LocalAttention",
]

__version__ = "0.0.1"

from .squeeze_excitation import SEAttention, MultiSEAttention

from .spatial_attention import SpatialAttention

from .channel_attention import ChannelAttention

from .conv_block_attention import ConvBlockAttention

from .efficient_channel_attention import EfficientChannelAttention

from .local_attention import LocalAttention



from typing import Optional

import torch
from torch import nn
import torch.nn.functional as F


class ChannelAttention(nn.Module):
    """
    The Channel Attention Module for Time Series (1D) or Image (2D) Analysis.
    This module focuses on 'what' is an informative feature by applying attention weights across channel dimensions.

    `Attention Here`: The linear layer is shared for both `max-pooled` and `avg-pooled` features.

    Reference: "CBAM: Convolutional Block Attention Module" by Sanghyun Woo, et al.

    URL: https://arxiv.org/abs/1807.06521
    """

    def __init__(
        self, n_dims: int, n_channels: int, reduction: Optional[int] = 4
    ) -> None:
        """
        Initialize the Channel Attention Module.

        :param n_dims: (int) The dimension of input data, either 1 (time series) or 2 (image).
        :param n_channels: (int) The number of input channels of time series or image
        :param reduction: (int) The reduction ratio for the intermediate layer in the channel attention block.
        """
        super().__init__()

        # Validate the input dimension
        assert n_dims in [1, 2], "The dimension of input data must be either 1 or 2."

        # Store parameters
        self.n_channels = n_channels
        self.reduction = reduction

        # Fully connected layers for the excitation operation
        self.linear = nn.Sequential(
            nn.Linear(
                in_features=self.n_channels,
                out_features=self.n_channels // self.reduction,
                bias=True,
            ),
            nn.ReLU(inplace=True),
            nn.Linear(
                in_features=self.n_channels // self.reduction,
                out_features=self.n_channels,
                bias=True,
            ),
        )

        # View shape for reshaping the excitation output
        self.view_shape = (1, 1) if n_dims == 2 else (1,)

        # The adaptive pooling layers for squeezing operation
        self.adaptive_max_pool = (
            nn.AdaptiveMaxPool2d(1) if n_dims == 2 else nn.AdaptiveMaxPool1d(1)
        )
        self.adaptive_avg_pool = (
            nn.AdaptiveAvgPool2d(1) if n_dims == 2 else nn.AdaptiveAvgPool1d(1)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass for the Channel Attention Module.

        :param x: (torch.Tensor)
                  1D Time Series: Input tensor of shape (batch_size, channels, seq_len);
                  2D Image: Input tensor of shape (batch_size, channels, height, width).

        :return: (torch.Tensor) Output tensor of the same shape as input.
        """
        # Compute max and average pooling along the channel dimension
        max = self.adaptive_max_pool(x)
        avg = self.adaptive_avg_pool(x)

        # Apply the fully connected layers for excitation operation
        batch_size, n_channels = x.size()[:2]
        linear_max = self.linear(max.view(batch_size, n_channels)).view(
            batch_size, n_channels, *self.view_shape
        )
        linear_avg = self.linear(avg.view(batch_size, n_channels)).view(
            batch_size, n_channels, *self.view_shape
        )

        # Combine and apply sigmoid activation to get attention map
        output = linear_max + linear_avg

        # Apply sigmoid activation and scale the input feature map
        output = F.sigmoid(output) * x

        return output



from typing import Optional

import torch
from torch import nn
import torch.nn.functional as F

from channel_attention.channel_attention import ChannelAttention
from channel_attention.spatial_attention import SpatialAttention


class ConvBlockAttention(nn.Module):
    """
    Convolutional Block Attention Module (CBAM) for Time Series (1D) or Image (2D) Analysis.
    This module sequentially applies Channel Attention and Spatial Attention to refine feature representations.

    Reference: "CBAM: Convolutional Block Attention Module" by Sanghyun Woo, et al.

    URL: https://arxiv.org/abs/1807.06521

    Also see: `ChannelAttention` and `SpatialAttention` classes.
    """

    def __init__(
        self,
        n_dims: int,
        n_channels: int,
        reduction: Optional[int] = 4,
        kernel_size: Optional[int] = 7,
    ) -> None:
        """
        Initialize the Convolutional Block Attention Module.

        :param n_dims: (int) The dimension of input data, either 1 (time series) or 2 (image).
        :param n_channels: (int) The number of input channels of time series or image.
        :param reduction: (int) The reduction ratio for the intermediate layer in the channel attention block.
        :param kernel_size: (int) The size of the convolutional kernel in the spatial attention block. Must be odd to maintain spatial dimensions.
        """
        super().__init__()

        # Initialize Channel Attention and Spatial Attention modules
        self.channel_attention = ChannelAttention(
            n_dims=n_dims, n_channels=n_channels, reduction=reduction
        )
        self.spatial_attention = SpatialAttention(
            n_dims=n_dims, kernel_size=kernel_size
        )

    def forward(self, x: torch.Tensor, with_residual: bool = True) -> torch.Tensor:
        """
        Forward pass for the Convolutional Block Attention Module.

        :param x: (torch.Tensor)
                  1D Time Series: Input tensor of shape (batch_size, channels, seq_len);
                  2D Image: Input tensor of shape (batch_size, channels, height, width).
        :param with_residual: (bool) Whether to include a residual connection from input to output.

        :return: (torch.Tensor) Output tensor of the same shape as input.
        """
        output = self.channel_attention(x)
        output = self.spatial_attention(output)

        if with_residual:
            return output + x
        return output



from typing import Optional, Union, Tuple

import torch
from torch import nn
import torch.nn.functional as F

from channel_attention.utils import create_conv_layer


class SoftPooling(torch.nn.Module):
    """
    The Soft Pooling Layer for Time Series (1D) or Image (2D) Analysis.
    This layer performs soft pooling by computing a weighted average of the input features,
    where the weights are determined by the exponential of the input features.
    """

    def __init__(
        self,
        n_dims: int,
        kernel_size: int,
        stride: Optional[int] = None,
        padding: Optional[int] = 0,
    ) -> None:
        """
        Initialize the SoftPooling layer.

        :param n_dims: (int) The dimension of input data, either 1 (time series) or 2 (image).
        :param kernel_size: (int) The size of the pooling kernel.
        :param stride: (int) The stride of the pooling operation. If None, it defaults to kernel_size.
        :param padding: (int) The amount of zero-padding added to both sides of the input.
        """
        super().__init__()
        # Store parameters
        self.n_dims = n_dims
        self.kernel_size = kernel_size

        # Create average pooling layer based on input dimension
        self.avg_pool = (
            nn.AvgPool2d(kernel_size, stride, padding, count_include_pad=False)
            if self.n_dims == 2
            else nn.AvgPool1d(kernel_size, stride, padding, count_include_pad=False)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass through the soft pooling layer.

        :param x: (torch.Tensor) Input tensor of shape (batch_size, channels, height, width) for 2D
                      or (batch_size, channels, seq_len) for 1D.

        :return: (torch.Tensor) Output tensor after soft pooling.
        """
        # Apply soft pooling operation
        x_exp = torch.exp(x)
        x_exp_pool = self.avg_pool(x_exp)

        # Get the weighted average
        x = self.avg_pool(x_exp * x)

        return x / x_exp_pool


class LocalModule(nn.Module):
    """
    Local Module for Channel Attention based on Local Importance.
    This module applies point-wise convolution, soft pooling, and additional convolutional layers
    to compute local attention weights for refining feature representations.

    Reference: "PlainUSR: Chasing Faster ConvNet for Efficient Super-Resolution"
    URL: https://arxiv.org/abs/2409.13435
    """

    def __init__(self, n_dims: int, n_channels: int, hidden_channels: int = 16) -> None:
        """
        Initialize the LocalModule.

        :param n_dims: (int) The dimension of input data, either 1 (time series) or 2 (image).
        :param n_channels: (int) The number of input channels of time series or image
        :param hidden_channels: (int) The number of hidden channels in the local attention module.
        """
        super().__init__()
        # Validate the input dimension
        assert n_dims in [1, 2], "The dimension of input data must be either 1 or 2."
        self.n_dims = n_dims

        # Store parameters
        self.n_channels = n_channels
        self.hidden_channels = hidden_channels

        # Create point-wise convolutional layer
        self.point_wise = create_conv_layer(
            n_dims=n_dims,
            in_channels=n_channels,
            out_channels=hidden_channels,
            kernel_size=1,
        )

        # Create soft pooling layer
        self.soft_pooling = SoftPooling(n_dims=n_dims, kernel_size=7, stride=3)

        # Create convolutional layers for local attention
        self.conv = nn.Sequential(
            create_conv_layer(
                n_dims=n_dims,
                in_channels=hidden_channels,
                out_channels=hidden_channels,
                kernel_size=3,
                stride=2,
                padding=1,
            ),
            create_conv_layer(
                n_dims=n_dims,
                in_channels=hidden_channels,
                out_channels=n_channels,
                kernel_size=3,
                stride=1,
                padding=1,
            ),
            # Gate activation
            nn.Sigmoid(),
        )

        self.mode = "linear" if n_dims == 1 else "bilinear"

    def get_data_size(self, data: torch.Tensor) -> Union[Tuple[int], Tuple[int, int]]:
        """
        Get the spatial dimensions of the input data.

        :param data: (torch.Tensor) Input tensor of shape (batch_size, n_channels, height, width) for 2D.

        :return: (Union[Tuple[int], Tuple[int, int]]) Spatial dimensions of the input data.
        """
        if self.n_dims == 2:
            return data.size(2), data.size(3)
        elif self.n_dims == 1:
            return (data.size(2),)
        else:
            raise ValueError("Invalid number of dimensions.")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass through the local attention module.

        :param x: (torch.Tensor) Input tensor of shape (batch_size, n_channels, height, width) for 2D or (batch_size, n_channels, seq_len) for 1D.

        :return: (torch.Tensor) Output tensor of shape (batch_size, n_channels, height, width) for 2D or (batch_size, n_channels, seq_len) for 1D.
        """
        # Forward pass through the local attention squeeze-and-excitation
        y = self.point_wise(x)
        y = self.soft_pooling(y)
        y = self.conv(y)

        # interpolate the heat map
        w = F.interpolate(
            input=y, size=self.get_data_size(x), mode=self.mode, align_corners=False
        )

        return w


class LocalModuleSpeed(nn.Module):
    """
    Local Module for Channel Attention based on Local Importance (Speed Version).
    This module applies convolutional layers to compute local attention weights for refining feature representations.
    This version is optimized for speed.
    """

    def __init__(self, n_dims: int, n_channels: int) -> None:
        """
        Initialize the LocalModuleSpeed.

        :param n_dims: (int) The dimension of input data, either 1 (time series) or 2 (image).
        :param n_channels: (int) The number of input channels of time series or image.
        """
        super().__init__()

        # Store parameters
        self.n_dims = n_dims
        self.n_channels = n_channels

        # Create convolutional layers for local attention
        self.conv = nn.Sequential(
            create_conv_layer(
                n_dims=n_dims,
                in_channels=n_channels,
                out_channels=n_channels,
                kernel_size=3,
                stride=1,
                padding=1,
            ),
            nn.Sigmoid(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass through the local attention module.

        :param x: (torch.Tensor) Input tensor of shape (batch_size, n_channels, height, width) for 2D or (batch_size, n_channels, seq_len) for 1D.

        :return: (torch.Tensor) Output tensor of shape (batch_size, n_channels, height, width) for 2D or (batch_size, n_channels, seq_len) for 1D.
        """
        # interpolate the heat map
        w = self.conv(x)

        return w


class LocalAttention(nn.Module):
    """
    Attention based on local importance for Time Series (1D) or Image (2D) Analysis.
    This module applies local attention weights to refine feature representations.

    Reference: "PlainUSR: Chasing Faster ConvNet for Efficient Super-Resolution"
    URL: https://arxiv.org/abs/2409.13435
    """

    def __init__(
        self,
        n_dims: int,
        n_channels: int,
        hidden_channels: int = 16,
        speed: Optional[bool] = False,
    ) -> None:
        """
        Initialize the LocalAttention module.

        :param n_dims: (int) The dimension of input data, either 1 (time series) or 2 (image).
        :param n_channels: (int) The number of input channels of time series or image
        :param hidden_channels: (int) The number of hidden channels in the local attention module.
        :param speed: (bool) Whether to use the speed-optimized version of the local attention module.
        """
        super().__init__()

        # Validate the input dimension
        assert n_dims in [1, 2], "The dimension of input data must be either 1 or 2."

        # Store parameters
        self.n_dims = n_dims
        self.n_channels = n_channels
        self.hidden_channels = hidden_channels
        self.speed = speed

        # Initialize the local attention body
        self.body = (
            LocalModule(
                n_dims=n_dims, n_channels=n_channels, hidden_channels=hidden_channels
            )
            if not speed
            else LocalModuleSpeed(n_dims=n_dims, n_channels=n_channels)
        )

        # Gate activation
        self.gate = nn.Sigmoid()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass through the local attention module.
        """
        # interpolate the heat map
        g = self.gate(x[:, :1])
        w = self.body(x=x)

        return x * w * g  # (w + g) #self.gate(x, w)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="channel-attention",
    packages=setuptools.find_packages(),
    version="0.0.1",
    description="A plug-and-play channel attention mechanism module implemented in PyTorch.",  # 包的简短描述
    url="https://github.com/wwhenxuan/Channel-Attention",
    author="whenxuan",
    author_email="wwhenxuan@gmail.com",
    keywords=[
        "Deep Learning",
        "Neural Networks",
        "Attention Mechanism",
        "Channel Attention",
        "PyTorch",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.6",
    install_requires=[
        "torch>=1.7.0",
    ],
)



import torch
from torch import nn


def create_conv_layer(
    n_dims: int,
    in_channels: int,
    out_channels: int,
    kernel_size: int,
    stride: int = 1,
    padding: int = 0,
    bias: bool = True,
    groups: int = 1,
) -> nn.Module:
    """Create a convolutional layer based on the number of dimensions."""
    if n_dims == 1:
        return nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=bias,
            groups=groups,
        )
    elif n_dims == 2:
        return nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=bias,
            groups=groups,
        )
    else:
        raise ValueError("Only 1D and 2D convolutional layers are supported.")



from typing import Optional

import torch
from torch import nn
import torch.nn.functional as F


class SpatialAttention(nn.Module):
    """
    The Spatial Attention Module for Time Series (1D) or Image (2D) Analysis.
    This module focuses on 'where' is an informative part by applying attention weights across spatial dimensions.

    Reference: "CBAM: Convolutional Block Attention Module" by Sanghyun Woo, et al.

    URL: https://arxiv.org/abs/1807.06521
    """

    def __init__(
        self, n_dims: int, kernel_size: Optional[int] = 7, bias: Optional[bool] = False
    ) -> None:
        """
        Initialize the Spatial Attention Module.

        :param n_dims: (int) The dimension of input data, either 1 (time series) or 2 (image).
        :param kernel_size: (int) The size of the convolutional kernel. Must be
                                odd to maintain the spatial dimensions.
        :param bias: (bool) Whether to include bias term in the convolutional layer.
        """
        super().__init__()

        # Validate the input dimension and kernel size
        assert n_dims in [1, 2], "The dimension of input data must be either 1 or 2."
        assert kernel_size % 2 == 1, "The kernel size must be odd."

        # Store parameters
        self.kernel_size = kernel_size
        self.stride = 1
        self.padding = kernel_size // 2
        self.bias = bias

        # parameters for 1D conv
        parameters = {
            "in_channels": 2,
            "out_channels": 1,
            "kernel_size": self.kernel_size,
            "stride": self.stride,
            "padding": self.padding,
            "dilation": 1,
            "bias": self.bias,
        }

        # Define convolutional layer based on input dimension
        self.conv = nn.Conv1d(**parameters) if n_dims == 1 else nn.Conv2d(**parameters)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass for the Spatial Attention Module.

        :param x: (torch.Tensor)
                  1D Time Series: Input tensor of shape (batch_size, channels, seq_len);
                  2D Image: Input tensor of shape (batch_size, channels, height, width).

        :return: (torch.Tensor) Output tensor of the same shape as input.
        """
        # Compute max and average pooling along the channel dimension
        max = torch.max(x, 1)[0].unsqueeze(1)
        avg = torch.mean(x, 1).unsqueeze(1)

        # Concatenate max and average pooled features
        concat = torch.cat((max, avg), dim=1)

        # Apply convolution and sigmoid activation to get attention map
        output = self.conv(concat)
        output = F.sigmoid(output) * x

        return output
